
local CRTRenderer = Object:extend()

function CRTRenderer:new()
  self.beam_color = {2.0 / 255, 254 / 255, 247 / 255}
  self.commands = {}
  self.x = 0
  self.y = 0
  self.beam_on = false
  self.scroll = 0
  self.canvas = love.graphics.newCanvas(love.graphics.getWidth(), love.graphics.getHeight(), {format="r32f"})
  self.canvas2 = love.graphics.newCanvas(love.graphics.getWidth(), love.graphics.getHeight())
  self.canvas3 = love.graphics.newCanvas(love.graphics.getWidth(), love.graphics.getHeight())

  self.bloom_canvas = love.graphics.newCanvas(love.graphics.getWidth() / 2, love.graphics.getHeight() / 2)
  self.bloom_canvas2 = love.graphics.newCanvas(love.graphics.getWidth() / 2, love.graphics.getHeight() / 2)

  self.intensity_map = love.graphics.newImage('intensity.png')
  self.intensity_map:setWrap('clamp', 'clamp')

  self.beam_shader = love.graphics.newShader [[
    uniform float intensity;
    uniform float intensity_mul;
    vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
      float i = Texel(texture, texture_coords).a;
      return vec4(intensity * intensity_mul * i, 0.0, 0.0, 1.0);
    }
  ]]

  self.draw_shader = love.graphics.newShader [[
    uniform vec3 beam_color;
    uniform Image intensity_texture;
    vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
      float i = Texel(texture, texture_coords).r;

      return Texel(intensity_texture, vec2(clamp(i* 0.5 + 0.25, 0.0, 1.0), 0.5));
    }
  ]]
  self.draw_shader:send('intensity_texture', self.intensity_map)

  self.darken_shader = love.graphics.newShader [[
    uniform float alpha;
    vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
      return vec4(alpha, alpha, alpha, alpha);
    }
  ]]

  self.threshold_shader = love.graphics.newShader [[
    vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
      float i = Texel(texture, texture_coords).r;
      float b = i >= 1.0 ? (i-1.0) : 0.0;
      return vec4(b,b,b,1.0);
    }
  ]]

  self.blur_shader = love.graphics.newShader [[
    uniform vec2 off;

    const int gauss_radius = 3;
    const float gauss_filter[gauss_radius] = float[gauss_radius](
      //0.06136, 0.24477, 0.38774, 0.24477, 0.06136

      0.27901, 0.44198, 0.27901
    );

    vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
      texture_coords = texture_coords - float(int(gauss_radius/2)) * off;
      vec4 out_color = vec4(0.0);
      for(int i = 0; i < gauss_radius; ++i) { 
        out_color += gauss_filter[i] * Texel(texture, texture_coords);
        texture_coords += off;
      }
      return out_color;
    }
  ]]

end


function CRTRenderer:add_command(cmd)
  table.insert(self.commands, cmd)
end


function CRTRenderer:goto(x, y)
  self:add_command {
    run = function (cmd, renderer)
      renderer.x = cmd.x
      renderer.y = cmd.y
    end,
    x = x,
    y = y
  }
end


function CRTRenderer:line(intensity, points)
  self:add_command {
    run = function (cmd, renderer)
      self.beam_shader:send('intensity', intensity)
      local pts = {}
      for i = 2, #points, 2 do
        pts[i] = math.floor(points[i] + self.scroll)
      end
      for i = 1, #points, 2 do
        pts[i] = points[i]
      end
      love.graphics.line(pts)
    end
  }
end


function CRTRenderer:update(dt)
  local n = table.maxn(self.commands)
  love.graphics.setCanvas(self.canvas)
  local alpha = 1 - dt*10
  self.darken_shader:send('alpha', alpha)
  love.graphics.setShader(self.darken_shader)
  love.graphics.setBlendMode('multiply', 'premultiplied')

  love.graphics.rectangle('fill', 0, 0, love.graphics.getWidth(), love.graphics.getHeight())

  love.graphics.setShader(self.beam_shader)
  love.graphics.setBlendMode('add')
  love.graphics.setLineJoin('bevel')
  love.graphics.setLineWidth(1)

  self.beam_shader:send('intensity_mul', dt * 60)
  for i = 1, n do
    local cmd = self.commands[i]
    self.commands[i] = nil
    cmd:run(self)
  end

  love.graphics.setBlendMode('replace')
  love.graphics.setCanvas(self.bloom_canvas)
  love.graphics.clear()
  love.graphics.setShader(self.threshold_shader)
  love.graphics.draw(self.canvas, 0, 0, 0, 0.5, 0.5)

  love.graphics.setShader(self.blur_shader)
  self.blur_shader:send('off', {1.0/256.0, 0.0})
  for i = 1, 3 do
    -- blur threshold layer to bc2
    love.graphics.setCanvas(self.bloom_canvas2)
    love.graphics.draw(self.bloom_canvas)

    -- blur bc2 to bc
    love.graphics.setCanvas(self.bloom_canvas)
    love.graphics.draw(self.bloom_canvas2)
  end
  self.blur_shader:send('off', {0.0, 1.0/256.0})
  for i = 1, 3 do
    -- blur threshold layer to bc2
    love.graphics.setCanvas(self.bloom_canvas2)
    love.graphics.draw(self.bloom_canvas)

    -- blur bc2 to bc
    love.graphics.setCanvas(self.bloom_canvas)
    love.graphics.draw(self.bloom_canvas2)
  end

  love.graphics.setCanvas()
end


function CRTRenderer:execute()

  love.graphics.setColor(255,255,255,255)
  love.graphics.setBlendMode('replace')

  love.graphics.setShader(self.draw_shader)
  love.graphics.setCanvas(self.canvas2)
  love.graphics.draw(self.canvas) -- turn intensity into color in c2

  -- blur v color from c2 to c3
  love.graphics.setShader(self.blur_shader)
  self.blur_shader:send('off', {0.0, 1.0/512.0})
  love.graphics.setCanvas(self.canvas3)
  love.graphics.draw(self.canvas2)

  -- blur h color from c3 to screen
  self.blur_shader:send('off', {1.0/512.0, 0.0})
  love.graphics.setBlendMode('add')
  love.graphics.setCanvas()
  love.graphics.draw(self.canvas3)

  --love.graphics.draw(self.canvas2)

  love.graphics.draw(self.bloom_canvas, 0, 0, 0, 2, 2)
--  love.graphics.setCanvas()
--  love.graphics.draw(self.canvas2)
end


function CRTRenderer:set_scroll(scroll)
  self.scroll = scroll
end


function CRTRenderer:print(font, intensity, text, x, y, w, align)
  self:add_command {
    run = function (cmd, renderer)
      self.beam_shader:send('intensity', intensity)
      love.graphics.setFont(font)
      love.graphics.printf(text, x, y, w, align)
    end
  }
end

function CRTRenderer:circle(intensity, x, y, r)
  self:add_command {
    run = function (cmd, renderer)
      self.beam_shader:send('intensity', intensity)
      love.graphics.circle('line', x, y + self.scroll, r)
    end
  }
end


return CRTRenderer
