--
-- lust
--
-- Copyright (c) 2014, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

--- Class for requesting preloading of sounds from different sources
local Sound = Object:extend()

--- Constructs new sound. Use Sound(...) to call it and obtain the 
--  created instance
--  @param filename The name of the sound file to load
--  @param type (optional, default='static') The memory type of the fail
--              as specified by love.audio.newSource
--  @param loop (optional, default=false)    Whether the sound shall be
--              played in a loop or not
function Sound:new(filename, type, loop)
  self.filename = filename
  self.type = type or 'static'
  self.loop = loop
end

--- Play the sound. If the sound is currently playing and not looping,
--  it will first be stopped. Playing the sound in multiple instances
--  at once is not yet supported.
function Sound:play()
  if not self.loop and self.source:isPlaying() then
    self:stop()
  end
  self.source:play()
end

--- Stop playing the sound. It is safe to call it on currently 
--  non-playing sounds
function Sound:stop()
  self.source:stop()
end

--- Internal function used to set the love.audio.Source of a sound
--  after it has been loaded and set the appropriate options 
--  @param source The loaded sound source
function Sound:_set_source(source)
  self.source = source
  if self.loop then
    self.source.setLooping(true)
  end
end

--- Global sound table. All sounds are available from this table 
--  after loading them.
--  So playing a sound boils down to lust.Sound.sounds.sound_name:play().
Sound.sounds = {}

--- Tell the sound subsystem that a sound will be required under a 
--  certain name. After calling Sound.load, the sound is available for
--  playback.
--  @param name     The name under which the sound will be available after 
--                  loading
--  @param filename The filename of the sound file to be loaded
function Sound.require(name, filename)
  Sound.sounds[name] = Sound(filename)
end

--- Load all sounds previously required by Sound.require
function Sound.load()
  for k,s in pairs(Sound.sounds) do
    s:_set_source(love.audio.newSource(s.filename, s.type))
  end
end

return Sound
