--
-- lust
--
-- Copyright (c) 2014, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

--- Simple input manager
local BoundInput = Object:extend()

--- Create new input manager. Call BoundInput(...) to call constructor and
--  obtain new instance.
function BoundInput:new()
  self.key_press_map   = {}
  self.key_release_map = {}

  self.action_map_func = function(key) return key end

  self.pressed = {}
end

--- Pass the key press event to a registered callback for that key, if one
--  exists.
--  @param key The key to be passed to the callback
--  @return true if a matching callback was found, false otherwise (useful for
--          chaining BoundInputs)
function BoundInput:dispatch_press(key)
  self.pressed[key] = key

  key = (key == 'escape' and 'escape') or self.action_map_func(key)
  local action = self.key_press_map[key]

  if action then
    action(key)
    return true
  end
  return false
end

--- Pass the key release event to a registered callback for that key, if one
--  exists.
--  @param key The key to be passed to the callback
--  @return true if a matching callback was found, false otherwise (useful for
--          chaining BoundInputs)
function BoundInput:dispatch_release(key)
  -- Make sure we don't get key release events that
  -- belong to other BoundInput's key press events
  if not self.pressed[key] then
    return false
  end

  self.pressed[key] = nil
  key = (key == 'escape' and 'escape') or self.action_map_func(key)
  local action = self.key_release_map[key]
    
  if action then
    action(key)
    return true
  end
  return false
end

--- Add new callback for key events.
--  @param key Keycode (as defined by love) to register the events for
--  @param press_cb (optional)   The callback to call when the key is pressed
--                               If the parameter is missing or nil, it will be
--                               removed from the callbacks
--  @param release_cb (optional) The callback to call when the key is pressed
--                               If the parameter is missing or nil, it will be
--                               removed from the callbacks
function BoundInput:bind_key(key, press_cb, release_cb)
  self.key_press_map[key] = press_cb
  self.key_release_map[key] = release_cb
end


function BoundInput:unbind_key(key)
  self.key_press_map[key] = nil
  self.key_release_map[key] = nil
end


local function NULLFUNC()
end


function BoundInput:bound_press_function(key)
  return self.key_press_map[key] or NULLFUNC
end


function BoundInput:bound_release_function(key)
  return self.key_release_map[key] or NULLFUNC
end


function BoundInput:is_key_bound(key)
  return self:bound_release_function(key) ~= nil or self:bound_press_function(key) ~= nil
end


function BoundInput:text_input(char)
end


function BoundInput:prepare()
  if not G.headless then
    love.keyboard.setKeyRepeat(false)
    love.keyboard.setTextInput(false)
  end
end

return BoundInput
