--
-- lust
--
-- Copyright (c) 2014, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

--- A composer for post processing effects. Will automatically manage canvases
--  and ping-pong between them, using one as input and one as output for the
--  post effects
local EffectComposer = Object:extend()

--- Create new EffectComposer. Call EffectComposer(...) to call the constructor
--  and obtain the new instance.
--  @param draw_function The function to call after the preparation of the
--                       drawing is completed. This can use any usual 
--                       love draw functions.
function EffectComposer:new(draw_function)
  self._post_effects = {}
  self._active_effects = {}
  self._canvases = { }
  self.draw_function = draw_function or function(cam) end
  for i = 1, 2 do
    self._canvases[i] = love.graphics.newCanvas(love.graphics.getWidth(),
                                                love.graphics.getHeight())
  end
end

--- Add a new effect to the effect chain
-- @param effect The effect instance to be added to the end of the chain.
--               Must implement PostEffect's interface.
function EffectComposer:add_post_effect(effect)
  self._post_effects[#self._post_effects+1] = effect
  self:_rebuild_active_list()
end

--- Enables or disables an effect. This rebuilds the list of active effects
--  and is therefore very expensive if many effects are registered in the
--  composer
--  @param effect The effect to be enabled or disabled
--  @param enable Whether the effect shall be included in the post effect
--                rendering chain or not
function EffectComposer:toggle_post_effect(effect, enable)
  effect.enabled = enable
  self:_rebuild_active_list()
end

--- Internal function used to rebuild the list of active effects.
function EffectComposer:_rebuild_active_list()
  for i = 1, #self._active_effects do
    self._active_effects[i] = nil
  end

  for i,e in ipairs(self._post_effects) do
    if e.enabled then
      self._active_effects[#self._active_effects+1] = e
    end
  end
end

--- Does all the drawing. Sets up the environment for drawing to the offscreen
--  canvas, calls the user defined draw function (as set in the constructor)
--  and passes the rendered image through the post effect chain.
--  @param ... Parameters passed to the user defined draw function
function EffectComposer:draw(...)
  love.graphics.reset()
  love.graphics.setShader()

  local effects = self._active_effects

  -- Draw directly to window if no effects are active
  if #effects == 0 then
    love.graphics.setCanvas()
    self.draw_function(...)
    return
  else
    love.graphics.setCanvas(self._canvases[1])
    self.draw_function(...)
  end


  -- Draw post effects using ping pong method between
  -- two offscreen canvases
  for i = 1,#effects-1 do
    local e = effects[i]
    love.graphics.reset()
    love.graphics.setCanvas(self._canvases[ i % 2 + 1 ])
    e:draw(self._canvases[(i+1)%2+1])
  end

  love.graphics.reset()
  love.graphics.setCanvas()
  effects[#effects]:draw(self._canvases[(#effects+1)%2+1])
end


return EffectComposer
