--
-- lust
--
-- Copyright (c) 2014, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

local Entity = Object:extend()

function Entity:new(state)
  self.state = state
  self.pos = { x = 0, y = 0 }
  self.vel = { x = 0, y = 0 }
  self.acc = { x = 0, y = 0 }
  self.lastpos = { x = 0, y = 0 }
  self.lastrotation = 0
  self.size = {x = 16, y = 16 }
  self.gravity = 0
  self.drag = { x = 200, y = 200 }
  self.bounce = 0
  self.rotation = 0

  self.scale = { x = 1, y = 1 }
  self.collisionOffset = { x = 0, y = 0 }

  self.vmax = 80

  self.layer = 0

  self.color = {r = 255,g = 255, b = 255, a = 255}
  self.offset = {x = 0, y = 0}
  self.hidden = false
  self.flickerfreq = 0
end


function Entity:sprite(filename, w, h, anims)
  local image = G.core.images:load(filename)
	if not self.anims then
		self.anims = {}
	end
  local iw, ih = image:getDimensions()
  local fpr = math.floor(iw / w)

  for _, a in ipairs(anims) do
    local f = {}
    for i, cf in ipairs(a[2]) do
      local x = cf % fpr
      local y = math.floor(cf / fpr)
      f[i] = G.core.quads:make(x * w, y * h, w, h, iw, ih)
    end

    self.anims[a[1]] = {
			image = image,
      frames = f,
      fps = a[3]
    }
  end

  self.size.x = w
  self.size.y = h
end


function Entity:updateGfx(dt)
  if self.anim then
    local oldframe = self.curframe
    self.curframe = math.fmod(self.curframe + dt * self.anim.fps, #self.anim.frames)

    if oldframe > self.curframe then
      if self.anim ~= self.nextAnim then
        self.anim = self.nextAnim
        self.curframe = 0
      end
    end
  end
end


function Entity:setNextAnim(anim)
  self.nextAnim = self.anims[anim]
end


function Entity:update(dt)
  self:updateGfx(dt)
  
  self.lastpos.x = self.pos.x
  self.lastpos.y = self.pos.y
  self.lastrotation = self.rotation

  local accy = self.acc.y + self.gravity
  local accx = self.acc.x

  if (math.abs(accx) < 0.1 or lust.tools.sig(accx) ~= lust.tools.sig(self.vel.x))  and self.drag.x > 0 then
    self.vel.x = self.vel.x - lust.tools.sig(self.vel.x) * math.min(self.drag.x*dt, math.abs(self.vel.x))
  end
  --else
  self.vel.x = self.vel.x + accx * dt
  --end

  if math.abs(accy) < 0.1 and self.drag.y > 0 then
    --self.vel.y = self.vel.y - lust.tools.sig(self.vel.y) * self.drag.y * dt
    self.vel.y = self.vel.y - lust.tools.sig(self.vel.y) * math.min(self.drag.y*dt, math.abs(self.vel.y))
  else
    self.vel.y = self.vel.y + accy * dt
  end

  self.pos.x = self.pos.x + dt * self.vel.x
  self.pos.y = self.pos.y + dt * self.vel.y

  if type(self.vmax) == "table" then
    if math.abs(self.vel.x) > self.vmax.x then
      self.vel.x = lust.tools.sig(self.vel.x) * self.vmax.x
    end
    if math.abs(self.vel.y) > self.vmax.y then
      self.vel.y = lust.tools.sig(self.vel.y) * self.vmax.y
    end
  else
    if math.sqrt(self.vel.x*self.vel.x + self.vel.y * self.vel.y) > self.vmax then
      self.vel.x, self.vel.y = lust.tools.scale(self.vel.x, self.vel.y, self.vmax)
    end
  end

  self:updateShape()
end


function Entity:play(anim, noreset)
  self.curAnim = anim
  self.anim = self.anims[anim]
  self.nextAnim = self.anim
  if not noreset or self.curframe > #self.anim.frames then
    self.curframe = 0
  end
end


function Entity:drawSelf(x, y, r)
  if self.anim then
    love.graphics.setColor(self.color.r, self.color.g, self.color.b, self.color.a)
    love.graphics.draw(self.anim.image, self.anim.frames[math.floor(self.curframe) + 1], math.floor(x), math.floor(y), r, self.scale.x, self.scale.y, self.offset.x, self.offset.y)
  end
end


function Entity:draw(lerp)
  if self.flickerfreq > 0 then
    self.flickertimer = (self.flickertimer + 1) % self.flickerfreq
    if self.flickertimer == 0 then
      self.hidden = not self.hidden
    end
  end
  if self.hidden then 
    return
  end
  local p = {
    x = lust.tools.lerp(self.lastpos.x, self.pos.x, lerp),
    y = lust.tools.lerp(self.lastpos.y, self.pos.y, lerp),
    r = lust.tools.lerp(self.lastrotation, self.rotation, lerp)
  }
  if self.color then
    love.graphics.setColor(self.color.r, self.color.g, self.color.b, self.color.a)
  else
    love.graphics.setColor(255,255,255,255)
  end
  self:drawSelf(math.floor(p.x), math.floor(p.y), p.r)

 --[[
  if self.shape then
    self.shape:draw()
  end

  love.graphics.point(self.pos.x, self.pos.y)
  --]]
end


function Entity:warpTo(x,y)
  self.pos = { x = x, y = y }
  self.lastpos = { x = x, y = y }
  self:updateShape()
end

function Entity:warp(x,y)
  self:warpTo(self.pos.x + x, self.pos.y + y)
end


function Entity:updateShape()
  if self.shape then
    self.shape:moveTo(self.pos.x + self.size.x / 2 - 0.5
                      + self.collisionOffset.x , -- * self.scale.x,
                      self.pos.y + self.size.y / 2 - 0.5
                       + self.collisionOffset.y) -- * self.scale.y)
  end
end


function Entity:collide(other)

end


function Entity:uncollide(other)

end


function Entity:kill()
  self.dead = true
end


function Entity:prepare()

end


function Entity:use()
end


function Entity:groundCollision()
  
end


function Entity:die()
  if self.shape then
    self.state.collider:remove(self.shape)
  end
end


function Entity:setProperty()
  -- do nothing :-(
end


function Entity:flicker(freq, duration)
  self.flickertimer = 0
  self.flickerfreq = freq
  duration = duration or 1
  self.state.timer:after(duration, function()
      self.flickerfreq = 0
      self.hidden = false
      return false
    end
  )
end
return Entity
