--
-- lust
--
-- Copyright (c) 2014, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

local Menu = Object:extend()
local BoundInput = require("lust.boundinput")

function Menu:new(core, screen, style)
  self.input = BoundInput()
  self.input:bind_key('up', function() self:inputPrev() end)
  self.input:bind_key('down', function() self:inputNext() end)
  self.input:bind_key('return', function() self:inputEnter() end)
  self.input:bind_key('escape', function() self:inputEscape() end)
  self.style = style
  self.core = core
  self.path = {screen}
  screen:enter()
end

function Menu:start()
  self.core.input:push(self.input)
end


function Menu:stop()
  self.core.input:pop()
end


function Menu:draw()
  if self.offset then
    love.graphics.translate(unpack(self.offset))
  end
  self.path[#self.path]:draw(self.style:newState())
end

function Menu:inputNext()
  self.path[#self.path]:inputNext(self)
end

function Menu:update(dt)
  self.path[#self.path]:update(dt)
end

function Menu:inputPrev()
  self.path[#self.path]:inputPrev(self)
end

function Menu:inputEnter()
  self.path[#self.path]:inputEnter(self)
end


function Menu:inputEscape()
  self:leaveScreen()
end


function Menu:enterScreen(screen)
  table.insert(self.path, screen)
  screen:enter(self)
end


function Menu:leaveScreen()
  if #self.path == 1 then
    (self.exitFunc or function() end)()
    return
  end
  self.path[#self.path]:leave()
  table.remove(self.path)
end

Menu.Screen = Object:extend()

function Menu.Screen:new(title, ...)
  self.title = title
  self.entries = { ... }
end


function Menu.Screen:getEntries()
  return self.entries
end


function Menu.Screen:update(dt)
end


function Menu.Screen:getTitle()
  return self.title
end

function Menu.Screen:drawEntries(style, entries)
  local maxw = 0
  for i,e in ipairs(entries) do
    if e.offset then
      style:offset(e.offset)
    end
    maxw = math.max(e:draw(style, self.selection == i), maxw)
  end
  return maxw
end


function Menu.Screen:draw(style)
  style:drawTitle(self:getTitle())
  return self:drawEntries(style, self:getEntries())
end


function Menu.Screen:inputNext(menu)
  self.selection = math.min(self.selection+1, #self:getEntries())
end

function Menu.Screen:inputPrev(menu)
  self.selection = math.max(self.selection-1, 1)
end

function Menu.Screen:inputEnter(menu)
  local entry = self:getEntries()[self.selection].dst
  if type(entry) == 'table' then
    menu:enterScreen(entry)
  elseif type(entry) == "nil" then
    menu:leaveScreen()
  elseif type(entry) == 'function' then
    entry(menu)
  end
end


function Menu.Screen:enter(menu)
  self.selection = 1
end


function Menu.Screen:leave()
end


Menu.Entry = Object:extend()

function Menu.Entry:new(caption, dst)
  self.caption = caption
  self.dst = dst
end


function Menu.Entry:draw(style, selected)
  return style:drawCaption(self.caption, selected)
end

return Menu
