--
-- lust
--
-- Copyright (c) 2014, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

local Tilemap = Object:extend()

local TilemapLayer = Object:extend()

function TilemapLayer:new(level, layer)
  local tw = level.map.tilewidth
  local th = level.map.tileheight
  local w = level.map.width
  local h = level.map.height
  self.level = level
  self.layer = layer
  self.rowsPerBatch = math.ceil(level.viewportHeight / th)
  self.colsPerBatch = math.ceil(level.viewportWidth / tw)
  self.batchRows = math.ceil(h / self.rowsPerBatch)
  self.batchCols = math.ceil(w / self.colsPerBatch)
  self.batches = {}
  for i = 0, self.batchRows * self.batchCols do
    self.batches[i] = love.graphics.newSpriteBatch(level.tiles, self.rowsPerBatch * self.colsPerBatch)
  end
  self:update()
  self.width  = w
  self.height = h
  self.tw = tw
  self.th = th
end


function TilemapLayer:update()
  local quads = self.level.quads
  local tw = self.level.map.tilewidth
  local th = self.level.map.tileheight
  for pr = 0, self.batchRows - 1 do
    for pc = 0, self.batchCols - 1 do
      local batchId = pr * self.batchCols + pc
      local batch = self.batches[pr * self.batchCols + pc]
      local baseCol = pc * self.colsPerBatch
      local baseRow = pr * self.rowsPerBatch
      batch:bind()
      batch:clear()
      for ir = 1, self.rowsPerBatch do
        for ic = 1, self.colsPerBatch do
          local r = baseRow + ir
          local c = baseCol + ic
          local idx = self:getTile(c,r)
          if idx ~= 0 then
            local q = quads[idx]
            batch:add(q, (c-1) * tw, (r-1)*th)
          end
        end
      end
      batch:unbind()
    end
  end

end


function TilemapLayer:getTile(x, y)
  if x < 1 or x > self.layer.width or y < 1 or y > self.layer.height then
    return 0
  else
    return self.layer.data[(y-1)*self.layer.width+x]
  end
end


function TilemapLayer:draw(camera)
  local r1 = math.floor(camera.y - camera.h / 2)
  local r2 = math.ceil(camera.y + camera.h / 2 )
  local c1 = math.floor(camera.x - camera.w / 2)
  local c2 = math.ceil(camera.x + camera.w / 2 )

  r1 = math.max(0, math.floor(r1 / (self.rowsPerBatch * self.level.map.tileheight)))
  r2 = math.min(self.batchRows-1, math.floor(r2 / (self.rowsPerBatch * self.level.map.tileheight)))
  c1 = math.max(0, math.floor(c1 / (self.colsPerBatch * self.level.map.tilewidth)))
  c2 = math.min(self.batchCols-1, math.floor(c2 / (self.colsPerBatch * self.level.map.tilewidth)))

  --print("drawing", c1, "to", c2, "and", r1, "to", r2)
  for c = c1, c2 do
    for r = r1, r2 do
      love.graphics.draw(self.batches[r * self.batchCols + c])
    end
  end
end

function TilemapLayer:drawBatch(batch)
  love.graphics.draw(self.batches[batch])
end


function Tilemap:new(state, mapfile, viewportWidth, viewportHeight, normalize)
  self.normalize = normalize or function(x) return "data/" .. x end
  self.state = state
  self.map = love.filesystem.load(self.normalize(mapfile))()
  self.core = state.core
  self.viewportWidth = viewportWidth or 50
  self.viewportHeight = viewportHeight or 50

  self.tileset = self.map.tilesets[1]

  local filename = self.tileset.image
  self.tiles = G.core.images:load(self.normalize(filename))

  local tilefilew = self.tiles:getWidth()
  local tilefileh = self.tiles:getHeight()

  local h = self.map.height
  local w = self.map.width
  local th = self.map.tileheight
  local tw = self.map.tilewidth
  local ntw = self.tiles:getWidth() / tw
  local nth = self.tiles:getHeight() / th

  self.tileheight = th
  self.tilewidth = tw
  self.width = w
  self.height = h

  self.quads = {}
  for r = 0, nth - 1 do
    for c = 0, ntw -1 do
      local q = G.core.quads:make(c * tw, r * th, tw, th, tilefilew, tilefileh)
      self.quads[r*ntw+c + 1] = q
    end
  end
  
  self.tilelayers = {}
  self.objectlayers = {}
  for _, e in pairs(self.map.layers) do
    if e.type == "tilelayer" then
      self.tilelayers[e.name] = TilemapLayer(self, e)
    elseif e.type == "objectgroup" then
      self.objectlayers[e.name] = e
    end
  end
end

return Tilemap
