--
-- lust
--
-- Copyright (c) 2014, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

-- TODO: To be partially replaced by lume. Probably some functionality can
--       also be moved into lume. Some functions have become useless over time.

local tools = {}

--- Return cube of a number
--  @param f The number to be cubed.
function tools.cu(f)
  return f*f*f
end

--- Return square of a number
--  @param f The number to be squared.
function tools.sq(f)
  return f*f
end

--- Linearly interpolate between two values
--  @param a The first value.
--  @param b The second value.
--  @param f The scaling factor between a and b. 0.0 returns value of a,
--           1.0 return value of b, values in-between returns interpolated
--           value. Values < 0 and > 1 are illegal and return nonsense.
function tools.lerp(a,b,f)
  return (1.0-f)*a+f*b
end

--- The signum function of a number.
--  @param x The number to compute the signum of.
--  @param epsilon The threshold for returning zero.
--- @return The signum of a value. Negative values are mapped to -1, positive
--  values to 1, values closer to 0 than epsilon are mapped to 0.
function tools.sig(x, epsilon)
  if x < 0 then
    return -1
  else
    return 1
  end
--[[
  if math.abs(x) < (epsilon or 0.00001) then
    return 0.0
  end
  return x / math.abs(x)
  ]]
end

--- Checks whether a value lies between to other values. The order of limits
--  does not matter.
--  @param a the value to check
--  @param b first limit (low or high)
--  @param b second limit (low or high)
--  @return true if a is beetween b and c (including), false otherwise.
function tools.between(a, b, c)
  -- Consider inaccuracies!
  return a >= math.min(b,c) - 0.00001 and a <= math.max(b,c) + 0.00001
end

function tools.ray_ray_intersection(x1,y1,rx1,ry1,x3,y3,rx3,ry3)
  local x2 = x1 + rx1
  local y2 = y1 + ry1
  local x4 = x3 + rx3
  local y4 = y3 + ry3
  local den = (x1 - x2)*(y3-y4) - (y1-y2)*(x3-x4)

  if math.abs(den) < 0.00001 then
    return nil
  end

  local px = ((x1*y2 - y1*x2)*(x3-x4) - (x1-x2)*(x3*y4 - y3*x4))/den
  local py = ((x1*y2 - y1*x2)*(y3-y4) - (y1-y2)*(x3*y4 - y3*x4))/den

  return px, py
end

--- Computes the intersection between two line segments
--  @return the intersection point, if any, nil otherwise
function tools.line_line_intersection(x1,y1,x2,y2,x3,y3,x4,y4)
  local den = (x1 - x2)*(y3-y4) - (y1-y2)*(x3-x4)

  if math.abs(den) < 0.00001 then
    return nil
  end

  local px = ((x1*y2 - y1*x2)*(x3-x4) - (x1-x2)*(x3*y4 - y3*x4))/den
  local py = ((x1*y2 - y1*x2)*(y3-y4) - (y1-y2)*(x3*y4 - y3*x4))/den

  -- TODO This functions very imprecise. Need to replace with proper bounds checking...
  if tools.between(px, x1, x2) and
     tools.between(px, x3, x4) and
     tools.between(py, y1, y2) and
     tools.between(py, y3, y4) then
    return px,py
  end

  return nil
end

--- Computes the intersection between a line segment and a ray
--  @return the intersection point, if any, nil otherwise
function tools.ray_line_intersection(x1,y1,rdx,rdy,x3,y3,x4,y4)
  local x2 = x1 + rdx
  local y2 = y1 + rdy
  local den = (x1 - x2)*(y3-y4) - (y1-y2)*(x3-x4)

  if math.abs(den) < 0.00001 then
    return nil
  end

  local px = ((x1*y2 - y1*x2)*(x3-x4) - (x1-x2)*(x3*y4 - y3*x4))/den
  local py = ((x1*y2 - y1*x2)*(y3-y4) - (y1-y2)*(x3*y4 - y3*x4))/den

  local gx = px - x3
  local gy = py - y3
  local lambda_sq = gx * (x4-x3) + gy * (y4-y3)
  local pos_ray = (px-x1) * rdx + (py-y1) * rdy

  if pos_ray < 0 then
    return nil
  end

  if lambda_sq >= 0.0 and tools.length(gx, gy) < tools.length(x4-x3, y4-y3) then
    return px, py
  end

  return nil
end

--- Rotate a point around the origin.
--  @param px X position of the point.
--  @param py Y position of the point.
--  @param r  The angle (in radians) to rotate the point around
--  @return x and y position of the rotated point
function tools.rotate(px, py, r)
  return math.cos(r) * px - math.sin(r) * py,
       math.sin(r) * px + math.cos(r) * py
end


--- Mirrors a polygon along the x axis (given as a list of vertices, as used by
--  love's drawPolygon and hardon colliders polygon shape), keeping the
--  correct counter clock wise winding order of the vertices.
--  @param shape The list of vertices to be mirrored.
--  @return The mirrored shape.
function tools.mirror_shape(shape)
  local out = {}
  for i = #shape-1, 1, -2 do
    out[#out+1] = -shape[i]
    out[#out+1] = shape[i+1]
  end
  return out
end

--- Compute length of a vector
--  @param x x component of vector
--  @param y y component of vector
--  @return length of vector
function tools.length(x,y)
  return math.sqrt(x*x+y*y)
end

--- Normalizes a two dimensional vector
--  @param x x component of vector
--  @param y y component of vector
--  @return normalized (length 1) vector
function tools.normalize(x,y)
  local len = tools.length(x, y)
  if len < 0.00001 then
    return 0, 0
  end
  local lendiv = 1.0 / len
  return x*lendiv,y*lendiv
end

--- Return one normal of given vector
--  @param x x component of vector
--  @param y y component of vector
--  @return normal of vector
function tools.normal(x,y)
  return y,-x
end

--- Scale vector to specified length
--  @param x x component of vector
--  @param y y component of vector
--  @param length the requested new length for the vector
--  @return The vector scaled to the requested length or 0,0 if
--          the input vector is 0,0
function tools.scale(x,y,length)
  if x == 0 and y == 0 then
    return 0,0
  end
  local factor = length / tools.length(x,y)
  return factor * x, factor * y
end


function tools.project_point_to_ray(px,py,rx,ry,rdx, rdy)
  local ldx, ldy = tools.normalize(rdx, rdy)
  local u = (px-rx)*ldx + (py-ry)*ldy
  return rx + ldx * u, ry + ldy * u
end


function tools.point_line_pivot(px,py,lx1,ly1,lx2,ly2)
  local ldx = lx2 - lx1
  local ldy = ly2 - ly1
  local u = (px-lx1)*ldx + (py-ly1)*ldy
  local llsq = ldx*ldx+ldy*ldy
  u = u / llsq
  if u >= 0 and u <= 1 then
    return lx1 + u * ldx, ly1 + u * ldy
  elseif u < 0 then
    return lx1, ly1
  else
    return lx2, ly2
  end
end


function tools.point_line_distance(px,py,lx1,ly1,lx2,ly2)
  local pivotx, pivoty = tools.point_line_pivot(px,py,lx1,ly1,lx2,ly2)
  return tools.length(pivotx - px, pivoty - py)
end


function tools.line_line_distance(l1x1, l1y1, l1x2, l1y2,
                                  l2x1, l2y1, l2x2, l2y2)

  local pld = tools.point_line_distance
  return math.min(
    pld(l1x1, l1y1, l2x1, l2y1, l2x2, l2y2),
    pld(l1x2, l1y2, l2x1, l2y1, l2x2, l2y2),
    pld(l2x1, l2y1, l1x1, l1y1, l1x2, l1y2),
    pld(l2x2, l2y2, l1x1, l1y1, l1x2, l1y2)
  )
end


-- Clip a line against a ray defined as a point and normal vector
-- normal vector is supposed to point to the inside of the polygon
function tools.clip_line(x1, y1, x2, y2, rx, ry, nx, ny)
  local d1 = (x1 - rx) * nx + (y1 - ry) * ny
  local d2 = (x2 - rx) * nx + (y2 - ry) * ny

  if d1 < 0 and d2 < 0 then
    return
  elseif d1 >= 0 and d2 >= 0 then
    return x1, y1, x2, y2
  end

  local ix, iy = tools.ray_ray_intersection(rx, ry, ny, -nx, x1, y1, x2-x1, y2-y1)
  if d1 < 0 then
    return ix, iy, x2, y2
  else
    return x1, y1, ix, iy
  end
end


function tools.scoped_set(table, key, value)
  local parts = {}
  for i in key:gmatch('[^.]+') do
    parts[#parts+1] = i
  end

  for i = 1, #parts-1 do
    table = table[parts[i]]
  end

  table[parts[#parts]] = value
end


function tools.scoped_get(table, key)
  local t2 = table
  for i in key:gmatch('[^.]+') do
    if table == nil then
      return
    end
    table = table[i]
  end
  if t2 == table then return nil end
  return table
end


function tools.point_in_polygon(x,y,polygon)
  -- Even-odd-rule point-in-polygon check
  local count = 0
  for i = 0, #polygon-1 do
    local point1 = polygon[i+1]
    local point2 = polygon[(i+1) % #polygon + 1]
    if tools.ray_line_intersection(x, y, 1, 0, point1[1], point1[2],
                                               point2[1], point2[2]) then
      count = count + 1
    end
  end
  return count % 2 == 1
end

return tools
